<?php

namespace Guzzle\Service\Description;

use Guzzle\Common\Exception\InvalidArgumentException;

/**
 * Data object holding the information of an API command
 */
class Operation implements OperationInterface {
	/** @var string Default command class to use when none is specified */
	const DEFAULT_COMMAND_CLASS = 'Guzzle\\Service\\Command\\OperationCommand';

	/** @var Parameter Additional parameters schema */
	protected $additionalParameters;

	/** @var string Class of the command object */
	protected $class;

	/** @var array Extra operation information */
	protected $data;

	/** @var bool Whether or not the command is deprecated */
	protected $deprecated;

	/** @var ServiceDescriptionInterface */
	protected $description;

	/** @var string Reference URL providing more information about the operation */
	protected $documentationUrl;

	/** @var array Array of errors that could occur when running the command */
	protected $errorResponses;

	/** @var string HTTP method */
	protected $httpMethod;

	/** @var string Name of the command */
	protected $name;

	/** @var string A longer text field to explain the behavior of the operation. */
	protected $notes;

	/** @var array Parameters */
	protected $parameters = [];

	/** @var array Hashmap of properties that can be specified. Represented as a hash to speed up constructor. */
	protected static $properties = [
		'name' => true, 'httpMethod' => true, 'uri' => true, 'class' => true, 'responseClass' => true,
		'responseType' => true, 'responseNotes' => true, 'notes' => true, 'summary' => true, 'documentationUrl' => true,
		'deprecated' => true, 'data' => true, 'parameters' => true, 'additionalParameters' => true,
		'errorResponses' => true,
	];

	/** @var string This is what is returned from the method */
	protected $responseClass;

	/** @var string Information about the response returned by the operation */
	protected $responseNotes;

	/** @var string Type information about the response */
	protected $responseType;

	/** @var string This is a short summary of what the operation does */
	protected $summary;

	/** @var string HTTP URI of the command */
	protected $uri;

	/**
	 * Builds an Operation object using an array of configuration data:
	 * - name:               (string) Name of the command
	 * - httpMethod:         (string) HTTP method of the operation
	 * - uri:                (string) URI template that can create a relative or absolute URL
	 * - class:              (string) Concrete class that implements this command
	 * - parameters:         (array) Associative array of parameters for the command. {@see Parameter} for information.
	 * - summary:            (string) This is a short summary of what the operation does
	 * - notes:              (string) A longer text field to explain the behavior of the operation.
	 * - documentationUrl:   (string) Reference URL providing more information about the operation
	 * - responseClass:      (string) This is what is returned from the method. Can be a primitive, PSR-0 compliant
	 *                       class name, or model.
	 * - responseNotes:      (string) Information about the response returned by the operation
	 * - responseType:       (string) One of 'primitive', 'class', 'model', or 'documentation'. If not specified, this
	 *                       value will be automatically inferred based on whether or not there is a model matching the
	 *                       name, if a matching PSR-0 compliant class name is found, or set to 'primitive' by default.
	 * - deprecated:         (bool) Set to true if this is a deprecated command
	 * - errorResponses:     (array) Errors that could occur when executing the command. Array of hashes, each with a
	 *                       'code' (the HTTP response code), 'reason' (response reason phrase or description of the
	 *                       error), and 'class' (a custom exception class that would be thrown if the error is
	 *                       encountered).
	 * - data:               (array) Any extra data that might be used to help build or serialize the operation
	 * - additionalParameters: (null|array) Parameter schema to use when an option is passed to the operation that is
	 *                                      not in the schema
	 *
	 * @param array                       $config      Array of configuration data
	 * @param ServiceDescriptionInterface $description Service description used to resolve models if $ref tags are found
	 */
	public function __construct(array $config = [], ServiceDescriptionInterface $description = null) {
		$this->description = $description;

		// Get the intersection of the available properties and properties set on the operation
		foreach (array_intersect_key($config, self::$properties) as $key => $value) {
			$this->{$key} = $value;
		}

		$this->class = $this->class ?: self::DEFAULT_COMMAND_CLASS;
		$this->deprecated = (bool) $this->deprecated;
		$this->errorResponses = $this->errorResponses ?: [];
		$this->data = $this->data ?: [];

		if (!$this->responseClass) {
			$this->responseClass = 'array';
			$this->responseType = 'primitive';
		} elseif ($this->responseType) {
			// Set the response type to perform validation
			$this->setResponseType($this->responseType);
		} else {
			// A response class was set and no response type was set, so guess what the type is
			$this->inferResponseType();
		}

		// Parameters need special handling when adding
		if ($this->parameters) {
			foreach ($this->parameters as $name => $param) {
				if ($param instanceof Parameter) {
					$param->setName($name)->setParent($this);
				} elseif (is_array($param)) {
					$param['name'] = $name;
					$this->addParam(new Parameter($param, $this->description));
				}
			}
		}

		if ($this->additionalParameters) {
			if ($this->additionalParameters instanceof Parameter) {
				$this->additionalParameters->setParent($this);
			} elseif (is_array($this->additionalParameters)) {
				$this->setadditionalParameters(new Parameter($this->additionalParameters, $this->description));
			}
		}
	}

	/**
	 * Add an error to the command
	 *
	 * @param string $code   HTTP response code
	 * @param string $reason HTTP response reason phrase or information about the error
	 * @param string $class  Exception class associated with the error
	 *
	 * @return self
	 */
	public function addErrorResponse($code, $reason, $class) {
		$this->errorResponses[] = ['code' => $code, 'reason' => $reason, 'class' => $class];

		return $this;
	}

	/**
	 * Add a parameter to the command
	 *
	 * @param Parameter $param Parameter to add
	 *
	 * @return self
	 */
	public function addParam(Parameter $param) {
		$this->parameters[$param->getName()] = $param;
		$param->setParent($this);

		return $this;
	}

	/**
	 * Get the additionalParameters of the operation
	 *
	 * @return Parameter|null
	 */
	public function getAdditionalParameters() {
		return $this->additionalParameters;
	}

	public function getClass() {
		return $this->class;
	}

	public function getData($name) {
		return isset($this->data[$name]) ? $this->data[$name] : null;
	}

	public function getDeprecated() {
		return $this->deprecated;
	}

	public function getDocumentationUrl() {
		return $this->documentationUrl;
	}

	public function getErrorResponses() {
		return $this->errorResponses;
	}

	public function getHttpMethod() {
		return $this->httpMethod;
	}

	public function getName() {
		return $this->name;
	}

	public function getNotes() {
		return $this->notes;
	}

	public function getParam($param) {
		return isset($this->parameters[$param]) ? $this->parameters[$param] : null;
	}

	public function getParamNames() {
		return array_keys($this->parameters);
	}

	public function getParams() {
		return $this->parameters;
	}

	public function getResponseClass() {
		return $this->responseClass;
	}

	public function getResponseNotes() {
		return $this->responseNotes;
	}

	public function getResponseType() {
		return $this->responseType;
	}

	public function getServiceDescription() {
		return $this->description;
	}

	public function getSummary() {
		return $this->summary;
	}

	public function getUri() {
		return $this->uri;
	}

	public function hasParam($name) {
		return isset($this->parameters[$name]);
	}

	/**
	 * Remove a parameter from the command
	 *
	 * @param string $name Name of the parameter to remove
	 *
	 * @return self
	 */
	public function removeParam($name) {
		unset($this->parameters[$name]);

		return $this;
	}

	/**
	 * Set the additionalParameters of the operation
	 *
	 * @param Parameter|null $parameter Parameter to set
	 *
	 * @return self
	 */
	public function setAdditionalParameters($parameter) {
		if ($this->additionalParameters = $parameter) {
			$this->additionalParameters->setParent($this);
		}

		return $this;
	}

	/**
	 * Set the concrete class of the command
	 *
	 * @param string $className Concrete class name
	 *
	 * @return self
	 */
	public function setClass($className) {
		$this->class = $className;

		return $this;
	}

	/**
	 * Set a particular data point on the operation
	 *
	 * @param string $name  Name of the data value
	 * @param mixed  $value Value to set
	 *
	 * @return self
	 */
	public function setData($name, $value) {
		$this->data[$name] = $value;

		return $this;
	}

	/**
	 * Set whether or not the command is deprecated
	 *
	 * @param bool $isDeprecated Set to true to mark as deprecated
	 *
	 * @return self
	 */
	public function setDeprecated($isDeprecated) {
		$this->deprecated = $isDeprecated;

		return $this;
	}

	/**
	 * Set the URL pointing to additional documentation on the command
	 *
	 * @param string $docUrl Documentation URL
	 *
	 * @return self
	 */
	public function setDocumentationUrl($docUrl) {
		$this->documentationUrl = $docUrl;

		return $this;
	}

	/**
	 * Set all of the error responses of the operation
	 *
	 * @param array $errorResponses Hash of error name to a hash containing a code, reason, class
	 *
	 * @return self
	 */
	public function setErrorResponses(array $errorResponses) {
		$this->errorResponses = $errorResponses;

		return $this;
	}

	/**
	 * Set the HTTP method of the command
	 *
	 * @param string $httpMethod Method to set
	 *
	 * @return self
	 */
	public function setHttpMethod($httpMethod) {
		$this->httpMethod = $httpMethod;

		return $this;
	}

	/**
	 * Set the name of the command
	 *
	 * @param string $name Name of the command
	 *
	 * @return self
	 */
	public function setName($name) {
		$this->name = $name;

		return $this;
	}

	/**
	 * Set a longer text field to explain the behavior of the operation.
	 *
	 * @param string $notes Notes on the operation
	 *
	 * @return self
	 */
	public function setNotes($notes) {
		$this->notes = $notes;

		return $this;
	}

	/**
	 * Set what is returned from the method. Can be a primitive, class name, or model. For example: 'array',
	 * 'Guzzle\\Foo\\Baz', or 'MyModelName' (to reference a model by ID).
	 *
	 * @param string $responseClass Type of response
	 *
	 * @return self
	 */
	public function setResponseClass($responseClass) {
		$this->responseClass = $responseClass;
		$this->inferResponseType();

		return $this;
	}

	/**
	 * Set notes about the response of the operation
	 *
	 * @param string $notes Response notes
	 *
	 * @return self
	 */
	public function setResponseNotes($notes) {
		$this->responseNotes = $notes;

		return $this;
	}

	/**
	 * Set qualifying information about the responseClass. One of 'primitive', 'class', 'model', or 'documentation'
	 *
	 * @param string $responseType Response type information
	 *
	 * @return self
	 * @throws InvalidArgumentException
	 */
	public function setResponseType($responseType) {
		static $types = [
			self::TYPE_PRIMITIVE => true,
			self::TYPE_CLASS => true,
			self::TYPE_MODEL => true,
			self::TYPE_DOCUMENTATION => true,
		];
		if (!isset($types[$responseType])) {
			throw new InvalidArgumentException('responseType must be one of ' . implode(', ', array_keys($types)));
		}

		$this->responseType = $responseType;

		return $this;
	}

	public function setServiceDescription(ServiceDescriptionInterface $description) {
		$this->description = $description;

		return $this;
	}

	/**
	 * Set a short summary of what the operation does
	 *
	 * @param string $summary Short summary of the operation
	 *
	 * @return self
	 */
	public function setSummary($summary) {
		$this->summary = $summary;

		return $this;
	}

	/**
	 * Set the URI template of the command
	 *
	 * @param string $uri URI template to set
	 *
	 * @return self
	 */
	public function setUri($uri) {
		$this->uri = $uri;

		return $this;
	}

	public function toArray() {
		$result = [];
		// Grab valid properties and filter out values that weren't set
		foreach (array_keys(self::$properties) as $check) {
			if ($value = $this->{$check}) {
				$result[$check] = $value;
			}
		}
		// Remove the name property
		unset($result['name']);
		// Parameters need to be converted to arrays
		$result['parameters'] = [];
		foreach ($this->parameters as $key => $param) {
			$result['parameters'][$key] = $param->toArray();
		}
		// Additional parameters need to be cast to an array
		if ($this->additionalParameters instanceof Parameter) {
			$result['additionalParameters'] = $this->additionalParameters->toArray();
		}

		return $result;
	}

	/**
	 * Infer the response type from the responseClass value
	 */
	protected function inferResponseType() {
		static $primitives = ['array' => 1, 'boolean' => 1, 'string' => 1, 'integer' => 1, '' => 1];
		if (isset($primitives[$this->responseClass])) {
			$this->responseType = self::TYPE_PRIMITIVE;
		} elseif ($this->description && $this->description->hasModel($this->responseClass)) {
			$this->responseType = self::TYPE_MODEL;
		} else {
			$this->responseType = self::TYPE_CLASS;
		}
	}
}
